import re 
import string
import random 

class PasswordValidation:
    def __init__(
        self,
        min: int=None,
        max: int=None,
        uppercase=False,
        lowercase=False,
        number=False,
        symbol=False,
    ):
        """
        Args:
            min (int, optional): The minimum length requirement for the password.
            max (int, optional): The maximum length requirement for the password.
            uppercase (bool, optional): Specifies whether the password must contain at least one uppercase letter. 
            lowercase (bool, optional): Specifies whether the password must contain at least one lowercase letter. 
            number (bool, optional): Specifies whether the password must contain at least one number. 
            symbol (bool, optional): Specifies whether the password must contain at least one symbol. 
        """
        self.min = min
        self.max = max
        self.uppercase = uppercase
        self.lowercase = lowercase
        self.number = number
        self.symbol = symbol
   
    def validate(self, password: str):
        """
        Validates the password.

        Args:
            password (str): The password string to be validated.

        Returns:
            list: List of all errors, list will be empty if theres no errors
            """
        self.password = password 
        
        errors = []

        if self.min is not None and not self._min():
            errors.append('short_pass')

        if self.max is not None and not self._max():
            errors.append('long_pass')

        if self.uppercase and not self._uppercase():
            errors.append('no_uppercase')

        if self.lowercase and not self._lowercase():
            errors.append('no_lowercase')
        
        if self.number and not self._number():
            errors.append('no_number')
            
        if self.symbol and not self._symbol():
            errors.append('no_symbol')

        return errors
    
    def generate(self, length: int = 12):
        """
        Args:
            length(int) - Length of password to generate (default = 12)
        """
        password = []

        while len(password) < length: 
            if self.uppercase:
                password.append(random.choice(string.ascii_uppercase))

            if self.lowercase: 
                password.append(random.choice(string.ascii_lowercase))

            if self.number: 
                password.append(str(random.randint(0, 9)))

            if self.symbol:
                password.append(random.choice(list(string.punctuation)))

        # since we will append bunch of stuff, we have to trim it
        # probably cheaper than to check every time to break early, or do anything else fancy
        while len(password) > length:
            password = password[:-1]
        
        random.shuffle(password) # shuffling so it breaks the pattern

        return ''.join(password)
    
    def _min(self):
        return len(self.password) >= self.min

    def _max(self):
        return len(self.password) <= self.max

    def _uppercase(self):
        return bool(re.search(r'[A-Z]', self.password))

    def _lowercase(self):
        return bool(re.search(r'[a-z]', self.password))

    def _number(self):
        return bool(re.search(r'[0-9]', self.password))
        
    def _symbol(self):
        return bool(re.search(r'[\W_]', self.password))
